@@ -17,6 +17,7 @@
 
 
 #include "../common.h"
+#include "test.h"
 #include <assert.h>
 #include <errno.h>
 #include <inttypes.h>
@@ -30,7 +31,6 @@
 #include "../plist.h"
 #include "../strbuf.h"
 #include "../util.h"
-#include "test.h"
 
 
 #define Exit_TRUE      0  /* Exit_SUCCESS */
@@ -668,5 +668,15 @@ enum filecmp compare_files(const wchar_t *left, const wchar_t *right)
 	return FC_SAME;
 }
 
+#if YASH_ENABLE_HELP
+const char test_help[] = Ngt(
+"evaluate a conditional expression"
+);
+const char test_syntax[] = Ngt(
+"\ttest expression\n"
+"\t[ expression ]\n"
+);
+#endif
+
 
 /* vim: set ts=8 sts=4 sw=4 noet tw=80: */
