@@ -21,6 +21,7 @@
 
 #include <stddef.h>
 #include "key.h"
+#include "../util.h"
 
 
 /* This structure describes an editing mode. */
@@ -54,6 +55,8 @@ extern void le_set_mode(le_mode_id_T id);
 
 extern int bindkey_builtin(int argc, void **argv)
     __attribute__((nonnull));
+extern const char bindkey_help[], bindkey_syntax[];
+extern const struct xgetopt_T bindkey_options[];
 
 
 /* Returns the mode ID of the specified mode. */
