@@ -20,6 +20,7 @@
 #define YASH_YASH_H
 
 #include <sys/types.h>
+#include "util.h"
 
 
 extern pid_t shell_pid, shell_pgid;
@@ -41,10 +42,15 @@ extern void exec_input(int fd, const char *name,
 
 extern _Bool nextforceexit;
 
+extern const struct xgetopt_T force_help_options[];
+
 extern int exit_builtin(int argc, void **argv)
     __attribute__((nonnull));
+extern const char exit_help[], exit_syntax[];
+
 extern int suspend_builtin(int argc, void **argv)
     __attribute__((nonnull));
+extern const char suspend_help[], suspend_syntax[];
 
 
 /* Exits the shell with the last exit status.
